@@ -100,6 +100,59 @@ class TeeImpl : Tee {
   bazel::windows::AutoHandle output2_;
 };
 
+// Buffered input stream (based on a Windows HANDLE) with peek-ahead support.
+//
+// This class uses two consecutive "pages" where it buffers data from the
+// underlying HANDLE (wrapped in an AutoHandle). Both pages are always loaded
+// with data until there's no more data to read.
+//
+// The "active" page is the one where the read cursor is pointing. The other
+// page is the next one to be read once the client moves the read cursor beyond
+// the end of the active page.
+//
+// The client advances the read cursor with Advance(). When the cursor reaches
+// the end of the active page, the other page becomes the active one (whose data
+// is already buffered), and the old active page is loaded with new data from
+// the underlying file.
+class IFStreamImpl : IFStream {
+ public:
+  // Creates a new IFStream.
+  //
+  // If successful, then takes ownership of the HANDLE in 'handle', and returns
+  // a new IFStream pointer. Otherwise leaves 'handle' alone and returns
+  // nullptr.
+  static IFStream* Create(bazel::windows::AutoHandle* handle,
+                          DWORD max_page_size = 0x100000 /* 1 MB */);
+
+  bool Get(uint8_t* result) const override;
+  bool Advance() override;
+
+ protected:
+  bool PeekN(DWORD n, uint8_t* result) const override;
+
+ private:
+  bazel::windows::AutoHandle handle_;
+  const std::unique_ptr<uint8_t[]> data_;
+  const DWORD max_page_size_;
+  DWORD page1_size_;
+  DWORD page2_size_;
+  DWORD page_end_;
+  DWORD read_pos_;
+
+  IFStreamImpl(bazel::windows::AutoHandle* handle,
+               std::unique_ptr<uint8_t[]>&& data, DWORD data_size,
+               DWORD max_page_size)
+      : handle_(handle),
+        data_(std::move(data)),
+        max_page_size_(max_page_size),
+        page1_size_(data_size > max_page_size ? max_page_size : data_size),
+        page2_size_(data_size > max_page_size ? data_size - max_page_size : 0),
+        read_pos_(0),
+        page_end_(page1_size_) {}
+
+  bool Page1Active() const { return read_pos_ < max_page_size_; }
+};
+
 // A lightweight path abstraction that stores a Unicode Windows path.
 //
 // The class allows extracting the underlying path as a (immutable) string so
@@ -622,8 +675,8 @@ bool OpenFileForWriting(const Path& path, bazel::windows::AutoHandle* result) {
 bool OpenExistingFileForRead(const Path& abs_path,
                              bazel::windows::AutoHandle* result) {
   HANDLE h = CreateFileW(AddUncPrefixMaybe(abs_path).c_str(), GENERIC_READ,
-                         FILE_SHARE_READ | FILE_SHARE_DELETE, NULL,
-                         OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
+                         FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
+                         NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
   if (h == INVALID_HANDLE_VALUE) {
     DWORD err = GetLastError();
     LogErrorWithArgAndValue(__LINE__, "Failed to open file", abs_path.Get(),
@@ -634,7 +687,7 @@ bool OpenExistingFileForRead(const Path& abs_path,
   return true;
 }
 
-bool TouchFile(const Path& path) {
+bool CreateEmptyFile(const Path& path) {
   bazel::windows::AutoHandle handle;
   return OpenFileForWriting(path, &handle);
 }
@@ -826,7 +879,7 @@ bool ExportXmlPath(const Path& cwd, Path* test_outerr, Path* xml_log) {
          // TODO(ulfjack): Update Gunit to accept XML_OUTPUT_FILE and drop the
          // GUNIT_OUTPUT env variable.
          SetEnv(L"GUNIT_OUTPUT", L"xml:" + unix_result) &&
-         CreateDirectories(xml_log->Dirname()) && TouchFile(*test_outerr);
+         CreateDirectories(xml_log->Dirname()) && CreateEmptyFile(*test_outerr);
 }
 
 devtools_ijar::u4 GetZipAttr(const FileInfo& info) {
@@ -1201,13 +1254,18 @@ bool StartSubprocess(const Path& path, const std::vector<const wchar_t*>& args,
     return true;
   } else {
     DWORD err = GetLastError();
-    LogErrorWithValue(__LINE__, "CreateProcessW failed", err);
+    LogErrorWithValue(
+        __LINE__,
+        (std::wstring(L"CreateProcessW failed (") + cmdline.get() + L")")
+            .c_str(),
+        err);
     return false;
   }
 }
 
-int WaitForSubprocess(HANDLE process) {
+int WaitForSubprocess(HANDLE process, LARGE_INTEGER* end_time) {
   DWORD result = WaitForSingleObject(process, INFINITE);
+  QueryPerformanceCounter(end_time);
   switch (result) {
     case WAIT_OBJECT_0: {
       DWORD exit_code;
@@ -1313,6 +1371,44 @@ bool ParseArgs(int argc, wchar_t** argv, Path* out_argv0,
   return true;
 }
 
+bool ParseXmlWriterArgs(int argc, wchar_t** argv, const Path& cwd,
+                        Path* out_test_log, Path* out_xml_log,
+                        Duration* out_duration, int* out_exit_code) {
+  if (argc < 5) {
+    LogError(__LINE__,
+             "Usage: $0 <test_output_path> <xml_log_path>"
+             " <duration_in_seconds> <exit_code>");
+    return false;
+  }
+  if (!out_test_log->Set(argv[1]) || out_test_log->Get().empty()) {
+    LogError(__LINE__, (std::wstring(L"Failed to parse test log path from \"") +
+                        argv[1] + L"\"")
+                           .c_str());
+    return false;
+  }
+  out_test_log->Absolutize(cwd);
+  if (!out_xml_log->Set(argv[2]) || out_xml_log->Get().empty()) {
+    LogError(__LINE__, (std::wstring(L"Failed to parse XML log path from \"") +
+                        argv[2] + L"\"")
+                           .c_str());
+    return false;
+  }
+  out_xml_log->Absolutize(cwd);
+  if (!out_duration->FromString(argv[3])) {
+    LogError(__LINE__, (std::wstring(L"Failed to parse test duration from \"") +
+                        argv[3] + L"\"")
+                           .c_str());
+    return false;
+  }
+  if (!ToInt(argv[4], out_exit_code)) {
+    LogError(__LINE__, (std::wstring(L"Failed to parse exit code from \"") +
+                        argv[4] + L"\"")
+                           .c_str());
+    return false;
+  }
+  return true;
+}
+
 bool TeeImpl::Create(bazel::windows::AutoHandle* input,
                      bazel::windows::AutoHandle* output1,
                      bazel::windows::AutoHandle* output2,
@@ -1356,8 +1452,7 @@ int RunSubprocess(const Path& test_path,
                            test_path.Get() + L"\"");
     return 1;
   }
-  int result = WaitForSubprocess(process);
-  QueryPerformanceCounter(&end);
+  int result = WaitForSubprocess(process, &end);
 
   QueryPerformanceFrequency(&freq);
   end.QuadPart -= start.QuadPart;
@@ -1539,32 +1634,53 @@ std::string CreateErrorTag(int exit_code) {
   }
 }
 
-bool CreateXmlLog(const Path& output, const Path& test_outerr,
-                  const Duration duration, const int exit_code) {
+bool ShouldCreateXml(const Path& xml_log, bool* result) {
+  *result = true;
+
+  DWORD attr = GetFileAttributesW(AddUncPrefixMaybe(xml_log).c_str());
+  if (attr != INVALID_FILE_ATTRIBUTES) {
+    // The XML file already exists, maybe the test framework wrote it.
+    // Leave the file alone.
+    *result = false;
+    return true;
+  }
+
   std::wstring split_xml_generation;
   if (!GetEnv(L"EXPERIMENTAL_SPLIT_XML_GENERATION", &split_xml_generation)) {
     LogError(__LINE__, "Failed to get %EXPERIMENTAL_SPLIT_XML_GENERATION%");
     return false;
   }
   if (split_xml_generation == L"1") {
-    // Bazel generates the test xml as a separate action.
+    // Bazel generates the test xml as a separate action, so we don't have to
+    // create it.
+    *result = false;
+  }
+
+  return true;
+}
+
+bool CreateXmlLog(const Path& output, const Path& test_outerr,
+                  const Duration duration, const int exit_code,
+                  const bool delete_afterwards) {
+  bool should_create_xml;
+  if (!ShouldCreateXml(output, &should_create_xml)) {
+    LogError(__LINE__,
+             (std::wstring(L"CreateXmlLog(") + output.Get() + L")").c_str());
+    return false;
+  }
+  if (!should_create_xml) {
     return true;
   }
 
-  Defer delete_test_outerr([test_outerr]() {
+  Defer delete_test_outerr([test_outerr, delete_afterwards]() {
     // Delete the test's outerr file after we have the XML file.
     // We don't care if this succeeds or not, because the outerr file is not a
     // declared output.
-    DeleteFileW(test_outerr.Get().c_str());
+    if (delete_afterwards) {
+      DeleteFileW(test_outerr.Get().c_str());
+    }
   });
 
-  DWORD attr = GetFileAttributesW(AddUncPrefixMaybe(output).c_str());
-  if (attr != INVALID_FILE_ATTRIBUTES) {
-    // The XML file already exists, maybe the test framework wrote it.
-    // Leave the file alone.
-    return true;
-  }
-
   std::wstring test_name;
   int errors = (exit_code == 0) ? 0 : 1;
   std::string error_msg = CreateErrorTag(exit_code);
@@ -1589,10 +1705,10 @@ bool CreateXmlLog(const Path& output, const Path& test_outerr,
   std::stringstream stm;
   stm << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
          "<testsuites>\n"
-         "    <testsuite name=\""
+         "<testsuite name=\""
       << acp_test_name << "\" tests=\"1\" failures=\"0\" errors=\"" << errors
       << "\">\n"
-         "    <testcase name=\""
+         "<testcase name=\""
       << acp_test_name << "\" status=\"run\" duration=\"" << duration.seconds
       << "\" time=\"" << duration.seconds << "\">" << error_msg
       << "</testcase>\n"
@@ -1656,6 +1772,100 @@ Path Path::Dirname() const {
   return result;
 }
 
+IFStream* IFStreamImpl::Create(bazel::windows::AutoHandle* handle,
+                               DWORD max_page_size) {
+  std::unique_ptr<uint8_t[]> data(new uint8_t[max_page_size * 2]);
+  DWORD read;
+  if (!ReadFile(*handle, data.get(), max_page_size * 2, &read, NULL)) {
+    DWORD err = GetLastError();
+    if (err == ERROR_BROKEN_PIPE) {
+      read = 0;
+    } else {
+      LogErrorWithValue(__LINE__, "Failed to read from file", err);
+      return nullptr;
+    }
+  }
+  return new IFStreamImpl(handle, std::move(data), read, max_page_size);
+}
+
+bool IFStreamImpl::Get(uint8_t* result) const {
+  if (read_pos_ < page_end_) {
+    *result = data_[read_pos_];
+    return true;
+  } else {
+    return false;
+  }
+}
+
+bool IFStreamImpl::Advance() {
+  if (read_pos_ + 1 < page_end_) {
+    read_pos_++;
+    return true;
+  }
+  const bool page1_was_active = Page1Active();
+  // The new page should have already been loaded when we started reading the
+  // current one (or it was filled by the Create method). Its size should only
+  // be zero if we reached EOF.
+  if ((page1_was_active && page2_size_ == 0) ||
+      (!page1_was_active && page1_size_ == 0)) {
+    return false;
+  }
+  // Overwrite the *active* page, because read_pos_ is about to move out of it
+  // and the current inactive page will be the new active one.
+  if (!ReadFile(handle_,
+                page1_was_active ? data_.get() : (data_.get() + max_page_size_),
+                max_page_size_, page1_was_active ? &page1_size_ : &page2_size_,
+                NULL)) {
+    DWORD err = GetLastError();
+    if (err == ERROR_BROKEN_PIPE) {
+      // The stream is reading from a pipe, and there's no more data.
+      if (page1_was_active) {
+        page1_size_ = 0;
+      } else {
+        page2_size_ = 0;
+      }
+    } else {
+      LogErrorWithValue(__LINE__, "Failed to read from file", err);
+      return false;
+    }
+  }
+  page_end_ = page1_was_active ? max_page_size_ + page2_size_ : page1_size_;
+  read_pos_ = page1_was_active ? max_page_size_ : 0;
+  return true;
+}
+
+bool IFStreamImpl::PeekN(DWORD n, uint8_t* result) const {
+  if (n > 3) {
+    // We only need to support peeking at up to 3 bytes. The theoretical upper
+    // limit is max_page_size_ * 2 - 1, because the buffer can hold at most
+    // max_page_size_ * 2 bytes of data and peeking starts at the next byte.
+    return false;
+  }
+
+  if (page_end_ - read_pos_ > n) {
+    // The current page has enough data we can peek at.
+    for (DWORD i = 0; i < n; ++i) {
+      result[i] = data_[read_pos_ + 1 + i];
+    }
+    return true;
+  }
+  DWORD required_from_next_page = n - (page_end_ - 1 - read_pos_);
+  // Check that the next page has enough data.
+  if ((Page1Active() && page2_size_ < required_from_next_page) ||
+      (!Page1Active() && page1_size_ < required_from_next_page)) {
+    // Pages are loaded eagerly by Advance(). The only way the next page's size
+    // can be zero is if we reached EOF.
+    return false;
+  }
+  for (DWORD i = 0, pos = read_pos_ + 1; i < n; ++i, ++pos) {
+    if (pos == page_end_) {
+      pos = Page1Active() ? max_page_size_ : 0;
+    }
+    result[i] = data_[pos];
+  }
+  return true;
+}
+
 }  // namespace
 
 void ZipEntryPaths::Create(const std::string& root,
@@ -1716,7 +1926,7 @@ void ZipEntryPaths::Create(const std::string& root,
   entry_path_ptrs_.get()[relative_paths.size()] = nullptr;
 }
 
-int Main(int argc, wchar_t** argv) {
+int TestWrapperMain(int argc, wchar_t** argv) {
   Path argv0;
   std::wstring test_path_arg;
   Path test_path, exec_root, srcdir, tmpdir, test_outerr, xml_log;
@@ -1737,7 +1947,7 @@ int Main(int argc, wchar_t** argv) {
 
   Duration test_duration;
   int result = RunSubprocess(test_path, args, test_outerr, &test_duration);
-  if (!CreateXmlLog(xml_log, test_outerr, test_duration, result) ||
+  if (!CreateXmlLog(xml_log, test_outerr, test_duration, result, true) ||
       !ArchiveUndeclaredOutputs(undecl) ||
       !CreateUndeclaredOutputsAnnotations(undecl.annotations_dir,
                                           undecl.annotations)) {
@@ -1746,6 +1956,21 @@ int Main(int argc, wchar_t** argv) {
   return result;
 }
 
+int XmlWriterMain(int argc, wchar_t** argv) {
+  Path cwd, test_outerr, test_xml_log;
+  Duration duration;
+  int exit_code = 0;
+
+  if (!GetCwd(&cwd) ||
+      !ParseXmlWriterArgs(argc, argv, cwd, &test_outerr, &test_xml_log,
+                          &duration, &exit_code) ||
+      !CreateXmlLog(test_xml_log, test_outerr, duration, exit_code, false)) {
+    return 1;
+  }
+
+  return 0;
+}
+
 namespace testing {
 
 bool TestOnly_GetEnv(const wchar_t* name, std::wstring* result) {
@@ -1824,6 +2049,11 @@ bool TestOnly_CdataEscapeAndAppend(const std::wstring& abs_input,
          CdataEscapeAndAppend(input_path, output);
 }
 
+IFStream* TestOnly_CreateIFStream(bazel::windows::AutoHandle* handle,
+                                  DWORD page_size) {
+  return IFStreamImpl::Create(handle, page_size);
+}
+
 }  // namespace testing
 }  // namespace test_wrapper
 }  // namespace tools
